# ============================================================================
### PREMIÈRE PARTIE : AVEC MISE SOUS FORME TRIANGULAIRE


# Calcule le noyau à droite de la matrice A dans le cas
# où A est triangulaire.
# Rem. On peut fournir ell, l'indice de la dernière ligne non-nulle
# s'il a déjà été calculé
def right_kernel_triangular(A, ell=None):
    n = A.nrows()
    if ell is None:
        ell = n
        while A[ell-1] == 0:
            ell -= 1
    F = A.base_ring()
    if ell == n:
        return zero_vector(F, n)
    T = identity_matrix(F, n)
    for i in range(ell):
        for j in range(i+1, n):
            alpha = A[i,j]/A[i,i]
            T.add_multiple_of_column(j, i, -alpha)
    return [ vector(v) for v in T.columns()[ell:] ]


# Calcule une solution particuliere de Tx = b dans le cas
# où T est triangulaire
def solve_triangular(T, b, ell=None):
    n = T.nrows()
    F = T.base_ring()
    x = [ F.zero() for i in range(n) ]
    if T.is_zero():
        if b.is_zero():
            return zero_vector(F, n), 0
        return None
    if ell is None:
        ell = n
        while T[ell-1] == 0:
            ell -= 1
    # if any(b[i] != 0 for i in range(ell, n)):
    if b[ell:n] != 0:
        return None
    for i in range(ell-1, -1, -1):
        x[i] = T[i,i]**(-1) * (b[i] - sum([T[i,j]*x[j] for j in range(i+1, n)]))
    x = vector(F, x)
    return x, ell




# effectue l'élimination gaussienne sur A (et en même temps sur b)
def gaussian_elimination_triangular(A, b):
    F = A.base_ring()
    n = A.nrows()
    T = copy(A)
    bp = matrix(b).transpose()
    U = identity_matrix(F, n)
    P = identity_matrix(F, n)
    i = 0
    ell = n
    while (i < ell):
        j = i
        while (j < n and T[i,j] == 0):
            j += 1
        if j == n:
            ell -= 1
            T.swap_rows(i, ell)
            bp.swap_rows(i, ell)
            U.swap_rows(i, ell)
        else:
            T.swap_columns(i, j)
            P.swap_columns(i, j)
            for k in range(i+1, ell):
                alpha = T[k,i]*(T[i,i]**(-1))
                T.add_multiple_of_row(k, i, -alpha)
                bp.add_multiple_of_row(k, i, -alpha)
                U.add_multiple_of_row(k, i, -alpha)
            i += 1
    return T, U, P, vector(bp), ell



# résout complètement le système Ax = b
def solve_system_triangular(A, b):
    F = A.base_ring()
    n = A.nrows()
    T, U, P, bp, ell = gaussian_elimination_triangular(A, b)
    res = solve_triangular(T, bp)
    if res is None:
        return None
    y, ell = res
    basis_kernel = right_kernel_triangular(T, ell)
    SOL = [ P*y, [P*u for u in basis_kernel] ]
    return SOL


















# ============================================================================
### DEUXIÈME PARTIE : AVEC MISE SOUS FORME ÉCHELONNÉE



# Calcule le noyau à droite de la matrice A dans le cas
# où A est échelonnée.
# Rem. On peut fournir ell, l'indice de la dernière ligne non-nulle
# s'il a déjà été calculé
def right_kernel_ref(A, ell=None):
    n = A.nrows()
    F = A.base_ring()
    if ell is None:
        ell = n-1
        while A[ell] == 0:
            ell -= 1
    T = identity_matrix(F, n)
    L_pivots = []
    pivot = 0
    for i in range(ell+1):
        # calcul de l'indice colonne du pivot
        while A[i][pivot] == 0:
            pivot += 1
        L_pivots.append(pivot)
        for j in range(pivot+1, n):
            alpha = A[i,j]/A[i,pivot]
            T.add_multiple_of_column(j, pivot, -alpha)

    return [ vector(T.columns()[i]) for i in range(n) if i not in L_pivots ]


# Calcule une solution particuliere de Ax = b dans le cas
# où A est echelonnee
def solve_ref(A, b, ell=None, liste_pivots=None):
    n = A.nrows()
    F = A.base_ring()
    x = [ F.zero() for i in range(n) ]
    if A.is_zero():
        if b.is_zero():
            return zero_vector(F, n), 0
        return None
    if ell is None:
        ell = n-1
        while A[ell] == 0:
            ell -= 1
    if b[ell+1:n] != 0:
        return None
    if liste_pivots is None:
        liste_pivots = []
        i = 0
        pivot = 0
        for i in range(ell+1):
            while A[i,pivot] == 0:
                pivot += 1
            liste_pivots.append(pivot)
    for i in range(ell, -1, -1):
        s = liste_pivots[i]
        x[s] = A[i,s]**(-1) * (b[i] - sum([A[i,j]*x[j] for j in range(s+1, n)]))
    x = vector(F, x)
    return x, ell



# effectue l'élimination gaussienne sur A
def gaussian_elimination_ref(A):
    F = A.base_ring()
    n = A.nrows()
    T = copy(A)
    U = identity_matrix(F, n)
    i, j = 0, 0
    while (i < n):
        while (j < n and T[i:,j] == 0):
            j += 1
        if (j != n):
            k = i
            while (k < n and T[k,j] == 0):
                k += 1
            if (k != n):
                T.swap_rows(i, k)
                U.swap_rows(i, k)
                for k in range(i+1, n):
                    alpha = T[k,j]*(T[i,j]**(-1))
                    T.add_multiple_of_row(k, i, -alpha)
                    U.add_multiple_of_row(k, i, -alpha)
        i += 1
    return T, U



# résout complètement le système Ax = b
def solve_system_ref(A, b):
    F = A.base_ring()
    n = A.nrows()
    T, U = gaussian_elimination_ref(A)
    Ub = U * b 
    res = solve_ref(T, Ub)
    if res is None:
        return None
    y, ell = res
    basis_kernel = right_kernel_ref(T, ell)
    SOL = [ y, [u for u in basis_kernel] ]
    return SOL

